import unittest
import numpy as np
import rf_linkbudget as rf


class TestCircuit(unittest.TestCase):


    def test_SourceN0(self):

        cr = rf.Circuit('Example')
        src = rf.Source("Source")
        sink = rf.Sink("Sink")

        src['out'] >> sink['in']

        # create callback function
        def cb_src(self, f, p):
            return {'f': f, 'p': p, 'Tn': rf.RFMath.T0}

        src['out'].regCallback(cb_src)

        cr.finalise()

        sim = cr.simulate(network=cr.net,
                        start=cr['Source'],
                        end=cr['Sink'],
                        freq=[0],
                        power=[0])

        sim.setNoiseBandwidth(1)

        assert sim.extractValues('Tn', freq=0, power=0)[1][0] == rf.RFMath.T0
        assert sim.extractValues('Tn', freq=0, power=0)[1][1] == rf.RFMath.T0

        assert sim.extractValues('n', freq=0, power=0)[1][0] == rf.RFMath.N0
        assert sim.extractValues('n', freq=0, power=0)[1][1] == rf.RFMath.N0

        assert sim.extractValues('NF', freq=0, power=0)[1][0] == 0
        assert sim.extractValues('NF', freq=0, power=0)[1][1] == 0

    def test_Lna1dB(self):

        cr = rf.Circuit('Example')
        src = rf.Source("Source")
        lna = rf.Amplifier("LNA",
                       Gain=20,
                       NF=1,
                       OP1dB=20,
                       OIP3=40)
        sink = rf.Sink("Sink")

        src['out'] >> lna['in']
        lna['out'] >> sink['in']

        # create callback function
        def cb_src(self, f, p):
            return {'f': f, 'p': p, 'Tn': rf.RFMath.T0}

        src['out'].regCallback(cb_src)

        cr.finalise()

        sim = cr.simulate(network=cr.net,
                        start=cr['Source'],
                        end=cr['Sink'],
                        freq=[0],
                        power=[0])

        sim.setNoiseBandwidth(1)

        assert sim.extractLastValues('n', freq=0, power=0) == (rf.RFMath.N0 +20 +1)
        assert np.round(sim.extractLastValues('NF', freq=0, power=0),decimals=4) == 1

    def test_Attenuator10dB(self):

        cr = rf.Circuit('Example')
        src = rf.Source("Source")
        att = rf.Attenuator("Attenuator",
                        Att=np.array([10])
                        )
        sink = rf.Sink("Sink")

        src['out'] >> att['in']
        att['out'] >> sink['in']

        # create callback function
        def cb_src(self, f, p):
            return {'f': f, 'p': p, 'Tn': rf.RFMath.T0}

        src['out'].regCallback(cb_src)

        cr.finalise()

        sim = cr.simulate(network=cr.net,
                        start=cr['Source'],
                        end=cr['Sink'],
                        freq=[0],
                        power=[0])

        sim.setNoiseBandwidth(1)

        assert sim.extractLastValues('n', freq=0, power=0) == rf.RFMath.N0
        assert np.round(sim.extractLastValues('NF', freq=0, power=0),decimals=4) == 10

if __name__ == '__main__':
    unittest.main()



import unittest
import numpy as np
import rf_linkbudget as rf


class TestComponents(unittest.TestCase):

    def setUp(self):
        rf.Circuit("CircuitName")

    def test_createAmplifier(self):
        rf.Amplifier("name",
                     Gain=[(0, 10)],
                     NF=0.5, OP1dB=25, OIP3=35)

    def test_createAttenuator_1(self):
        rf.Attenuator("name",
                      Att=[0])

    def test_createAttenuator_2(self):
        rf.Attenuator("name",
                      Att=np.arange(1, 6, 1))

    def test_createAttenuator_3(self):
        rf.Attenuator("name",
                      Att=np.arange(1, 6, 1),
                      OP1dB=25, IIP3=35)

    def test_Filter(self):
        rf.Filter("name",
                  Att=[(0, 2.0)])

    def test_SPDT(self):
        rf.SPDT("name",
                Att=0.3)

    def test_Source(self):
        rf.Source("Source")

    def test_Sink(self):
        rf.Sink("Sink")


if __name__ == '__main__':
    unittest.main()



import numpy as np
from collections import defaultdict
import functools
import operator
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib import cm
from matplotlib.text import Text
from mpl_toolkits.mplot3d import Axes3D

# create docs
# sphinx-apidoc -F -o docs rf_linkbudget/

try:
    from .rfmath import RFMath
except ImportError:
    from rfmath import RFMath


class simResult():
    """
    | A class which gets generated by Circuit.simulate(...)
    | It hold convenience functions for plotting and accessing the right datsets

    Parameters
    ----------
    network : str
        circuit network
    data : dictionary
        circuit data, from a simulation
    **kwargs : kwargs
        all other kwargs gets saved in a variable with the same name
    """

    def __init__(self, network, data, **kwargs):
        """
        Parameters
        ----------
        network : str
            circuit network
        data : dictionary
            circuit data, from a simulation
        **kwargs : kwargs
            all other kwargs gets saved in a variable with the same name
        """
        self.network = network
        self.nodes = set()  # all different nodes in network
        self.data = data
        self.freq = [key for key in data]
        self.power = [key for key in data[self.freq[0]]]

        self.graphes = set(functools.reduce(operator.iconcat, [list(network[x][y].nodes) for x in network for y in network[x]], []))
        self.nodes = set([g.parent.name for g in self.graphes])

        for key, value in kwargs.items():
            setattr(self, key, value)

        dat = defaultdict(dict)  # convert dictionarys to pandas DataFrame
        for freq in data:
            for pwr in data[freq]:
                dat[pwr][freq] = pd.DataFrame(data[freq][pwr])

        self.pandas = pd.DataFrame(dat)

    @property
    def param(self):
        """
        | access the parameters as a list

        Returns
        ----------
        parameter : list
            list of parameters
        """
        tmp = self.pandas.applymap(lambda x: list(x.index))
        return set(functools.reduce(operator.concat, functools.reduce(operator.concat, tmp.values.tolist())))

    def __repr__(self):
        out = 'simulation result:\n'
        out += '@ frequencies : {}\n'.format(self.freq)
        out += '@ input power : {}\n'.format(self.power)
        return out

    def setNoiseBandwidth(self, B):
        """
        | set the noisebandwidth
        | changes the following parameters

            - n : noise power
            - SNR : signal to noise ratio
            - DYN : dynamic
        """

        df = self.pandas

        def setNoise(x):
            for i in range(len(x.loc['n'])):
                tn = x.loc['Tn'].iat[i]
                p = x.loc['p'].iat[i]
                sfdr = x.loc['SFDR'].iat[i]

                n = RFMath.convert_T_n(tn, B)
                snr = RFMath.calc_SNR(p, n)
                dyn = RFMath.calc_Dynamic(snr, sfdr)

                x.loc['n'].iat[i] = n
                x.loc['SNR'].iat[i] = snr
                x.loc['DYN'].iat[i] = dyn

            return x

        df = df.applymap(lambda x: setNoise(x))
        self.pandas = df

    def extractValues(self, param, freq, power, keys=None):
        """
        | convencience function to access the data in a structured way

        Parameters
        ----------
        param : str or list of str
            a single parameter or a list of parameters which shall be extracted
        freq : float or list of floats
            a single frequency or a list of frequencies which shall be extracted
        power : float or list of floats
            a single power value or a list of power values which shall be extracted
        keys : str or list of str, optional
            a single device port or a list of device ports.
            only these values are extracted.

        Returns
        ----------
        params : (ports, data)
            a tuple with all ports and the corresponding data
        """
        df = self.pandas.loc[freq][power]
        val = df.loc[param].values
        ports = list(df.loc[param].index)

        if keys is not None:
            ports = keys
            val = df.loc[param].reindex(keys, fill_value=None)

        return (ports, val)

    def extractLastValues(self, key, freq=None, power=None):
        """
        | convencience function to access the data at the last device port.
        | an easy way to export all "end" values to compare systems with each other

        Parameters
        ----------
        key : str
            a single device port or a list of device ports.
            only these values are extracted.
        freq : float, optional
            a single frequency or a list of frequencies which shall be extracted
        power : float, optional
            a single power value or a list of power values which shall be extracted

        Returns
        ----------
        params : pandas.DataFrame
            data
        """
        df = self.pandas
        df = df.applymap(lambda x: x.iloc[:, -1][key])

        if freq is not None and power is None:
            df = df.loc[freq][:]
        elif freq is None and power is not None:
            df = df.loc[:][power]
        elif freq is not None and power is not None:
            df = df.loc[freq][power]
        else:
            pass

        return df

    def plot_chain(self, param=None, freq=None, power=None, keys=None):
        """
        | plots a window with responsive fields which can be switched between (click or mousewheel)
        | plot_chain plots all data from the start to the end, by all ports between
        | we can select which ports to show, be defining the keys
        | when param, freq or power is not defined, all parameters will be shown

        Parameters
        ----------
        param : str or list of str, optional
            a single parameter or a list of parameters which shall be plotted
        freq : float or list of floats, optional
            a single frequency or a list of frequencies which shall be plotted
        power : float or list of floats, optional
            a single power value or a list of power values which shall be plotted
        keys : str or list of str, optional
            a single device port or a list of device ports.
            only these values are extracted.

        Notes
        -----
        return handler must remain, otherwise the plot will not be responsive anymore

        Returns
        ----------
        handler : callback handlers
        """
        freq = self.freq if freq is None else freq
        power = self.power if power is None else power
        param = self.param if param is None else param
        freq = freq if type(freq) is list else [freq]
        power = power if type(power) is list else [power]
        param = param if type(param) is set() else list(param)
        param = param if type(param) is list else [param]

        fig, ax = plt.subplots()
        fig.subplots_adjust(bottom=0.2)
        l, = plt.plot([], [])
        plt.xticks(rotation=45)
        ax.grid(True)

        bbox_props = dict(boxstyle="round", fc="w", ec="0.5", alpha=0.9, )
        tb = [ax.text(1.015, 0.95, param[0], picker=True, bbox=bbox_props, horizontalalignment='left', rotation=0, transform=ax.transAxes),
              ax.text(1.015, 0.85, freq[0], picker=True, bbox=bbox_props, horizontalalignment='left', rotation=0, transform=ax.transAxes),
              ax.text(1.015, 0.75, power[0], picker=True, bbox=bbox_props, horizontalalignment='left', rotation=0, transform=ax.transAxes)]

        class Index(object):
            v = {'param': param, 'freq': freq, 'power': power}
            i = dict(zip(['param', 'freq', 'power'], [0, 0, 0]))
            t = dict(zip(['param', 'freq', 'power'], tb))
            ref = {'fig': fig, 'plot': l, 'axis': ax}
            parent = self

            def _cb_rotate(self, event, direction, idx):
                self.i[idx] = np.mod(self.i[idx] + direction, len(self.v[idx]))
                self.redraw()

            def redraw(self):
                cparam, cfreq, cpower = [self.v['param'][self.i['param']], self.v['freq'][self.i['freq']], self.v['power'][self.i['power']]]
                xt, val = self.parent.extractValues(cparam, cfreq, cpower, keys=keys)
                self.t['param'].set_text(cparam)
                self.t['freq'].set_text('{:.1f}MHz'.format(cfreq/1e6))
                self.t['power'].set_text(cpower)
                self.ref['plot'].set_data(range(len(xt)), val)
                self.ref['axis'].set_xticks(range(len(xt)))
                self.ref['axis'].set_xticklabels(xt)
                self.ref['axis'].set_ylabel(cparam)
                self.ref['axis'].set_title("Performance: freq={:.1f}MHz, Pin={:.1f}dBm".format(cfreq/1e6, cpower), picker=True)
                self.ref['axis'].relim()
                self.ref['axis'].autoscale_view()
                self.ref['fig'].canvas.draw_idle()

        callback = Index()
        callback.redraw()

        def onpick1(event):
            if isinstance(event.artist, Text):
                text = event.artist
                direction = 1 if event.mouseevent.button == 1 or event.mouseevent.button == 'up' else -1
                if callback.t['param'] == text:
                    callback._cb_rotate(event=event, direction=direction, idx='param')
                elif callback.t['freq'] == text:
                    callback._cb_rotate(event=event, direction=direction, idx='freq')
                elif callback.t['power'] == text:
                    callback._cb_rotate(event=event, direction=direction, idx='power')

        cid = fig.canvas.mpl_connect('pick_event', onpick1)
        handler = (cid, callback)
        return handler

    def plot_total(self, param=None, freq=None, power=None):
        """
        | plots a window with responsive fields which can be switched between (click or mousewheel)
        | plot_total plots all "and" data in respect to frequency or power
        | we can select which ports to show, be defining the keys
        | when param, freq or power are not defined, all parameters will be shown

        Parameters
        ----------
        param : str or list of str, optional
            a single parameter or a list of parameters which shall be plotted
        freq : float or list of floats, optional
            a single frequency or a list of frequencies which shall be plotted
        power : float or list of floats, optional
            a single power value or a list of power values which shall be plotted

        Notes
        -----
        return handler must remain, otherwise the plot will not be responsive anymore

        Returns
        ----------
        handler : callback handlers
        """
        freq = self.freq if freq is None else freq
        power = self.power if power is None else power
        param = self.param if param is None else param
        freq = freq if type(freq) is list else [freq]
        power = power if type(power) is list else [power]
        param = param if type(param) is set() else list(param)
        param = param if type(param) is list else [param]

        if len(freq) == 1:
            varDependend = 'freq'
        elif len(power) == 1:
            varDependend = 'power'
        else:
            varDependend = 'freq'

        fig, ax = plt.subplots()
        fig.subplots_adjust(bottom=0.2)
        l, = plt.plot([], [])
        ax.grid(True)

        bbox_props = dict(boxstyle="round", fc="w", ec="0.5", alpha=0.9, )
        tb = [ax.text(1.015, 0.95, param[0], picker=True, bbox=bbox_props, horizontalalignment='left', rotation=0, transform=ax.transAxes),
              ax.text(1.015, 0.85, freq[0], picker=True, bbox=bbox_props, horizontalalignment='left', rotation=0, transform=ax.transAxes),
              ax.text(1.015, 0.75, power[0], picker=True, bbox=bbox_props, horizontalalignment='left', rotation=0, transform=ax.transAxes)]

        class Index(object):
            v = {'param': param, 'freq': freq, 'power': power}
            i = dict(zip(['param', 'freq', 'power'], [0, 0, 0]))
            t = dict(zip(['param', 'freq', 'power'], tb))
            ref = {'fig': fig, 'plot': l, 'axis': ax}
            varDep = varDependend
            parent = self

            def _cb_rotate(self, event, direction, idx):
                self.i[idx] = np.mod(self.i[idx] + direction, len(self.v[idx]))
                self.redraw()

            def redraw(self):
                cparam, cfreq, cpower = [self.v['param'][self.i['param']], self.v['freq'][self.i['freq']], self.v['power']] if self.varDep == 'freq' else [self.v['param'][self.i['param']], self.v['freq'], self.v['power'][self.i['power']]]
                df = self.parent.extractLastValues(cparam, cfreq, cpower)
                self.t['param'].set_text(cparam)
                self.t['freq'].set_text('{:.1f}MHz'.format(cfreq/1e6)) if self.varDep == 'freq' else self.t['freq'].set_text('freq')
                self.t['power'].set_text('power') if self.varDep == 'freq' else self.t['power'].set_text(cpower)
                self.ref['plot'].set_data(list(df.index), list(df.values))
                self.ref['axis'].set_ylabel(cparam)
                self.ref['axis'].set_title("Concat {} plot".format(cparam))
                self.ref['axis'].relim()
                self.ref['axis'].autoscale_view()
                self.ref['fig'].canvas.draw_idle()

            def swapDependendVar(self):
                if self.varDep == 'freq':
                    self.varDep = 'power'
                else:
                    self.varDep = 'freq'
                self.redraw()

        callback = Index()
        callback.redraw()

        def onpick1(event):
            if isinstance(event.artist, Text):
                text = event.artist
                direction = 1 if event.mouseevent.button == 1 or event.mouseevent.button == 'up' else -1

                if event.mouseevent.button == 1:
                    # clicked on item
                    if callback.t['freq'] == text or callback.t['power'] == text:
                        callback.swapDependendVar()
                else:
                    if callback.t['param'] == text:
                        callback._cb_rotate(event=event, direction=direction, idx='param')
                    elif callback.t['freq'] == text:
                        callback._cb_rotate(event=event, direction=direction, idx='freq')
                    elif callback.t['power'] == text:
                        callback._cb_rotate(event=event, direction=direction, idx='power')
                    else:
                        pass

        cid = fig.canvas.mpl_connect('pick_event', onpick1)
        handler = (cid, callback)
        return handler

    def plot_total_simple(self, param, freq=None, power=None, axis=None):
        """
        | plots a window with the defined parameter
        | plot_total_simple is a simpler variant of the other plots
        | it is not a responvice view, but it is easy to show different parameters at once

        Parameters
        ----------
        param : str or list of str
            a single parameter or a list of parameters which shall be plotted
        freq : float or list of floats, optional
            a single frequency or a list of frequencies which shall be plotted
        power : float or list of floats, optional
            a single power value or a list of power values which shall be plotted
        axis : axis object, optional
            currently not in use

        Notes
        -----
        return handler must remain, otherwise the plot will not be responsive anymore

        Examples
        ---------
        >>> t = sim1.plot_total_simple(['SFDR', 'SNR', 'DYN'], freq=430e6)
        >>> plt.show()

        Returns
        ----------
        handler : callback handlers
        """
        # Todo, use axis parameter to show different plots in one...
        fig, axis = plt.subplots()

        for p in param:
            df = self.extractLastValues(p, freq, power)

            if freq is not None and power is None:
                plt.plot(df, label=p)
                axis.set_xlabel('input frequency [MHz]')
            elif freq is None and power is not None:
                plt.plot(df.index/1e6, df.values, label=p)
                axis.set_xlabel('input power [dBm]')
            elif freq is not None and power is not None:
                raise ValueError('is a scalar')
            else:
                raise ValueError('two dimensional plot')
        # axis.set_ylabel('Values')
        plt.xticks(rotation=45)
        plt.xticks(df.index, df.index)
        plt.legend()
        fig.set_tight_layout({"pad": .0})
        plt.grid(True)

    def plot_surface(self, param=None, freq=None, power=None, keys=None):
        """
        | plots a window with responsive fields which can be switched between (click or mousewheel)
        | plot_surface plots a surface plot with all data in respect to al ports and power or frequency
        | we can select which ports to show, be defining the keys
        | when param, freq or power are not defined, all parameters will be shown

        Parameters
        ----------
        param : str or list of str, optional
            a single parameter or a list of parameters which shall be plotted
        freq : float or list of floats, optional
            a single frequency or a list of frequencies which shall be plotted
        power : float or list of floats, optional
            a single power value or a list of power values which shall be plotted
        keys : str or list of str, optional
            a single device port or a list of device ports.
            only these values are extracted.

        Notes
        -----
        return handler must remain, otherwise the plot will not be responsive anymore

        Returns
        ----------
        handler : callback handlers
        """
        freq = self.freq if freq is None else freq
        power = self.power if power is None else power
        param = self.param if param is None else param
        freq = freq if type(freq) is list else [freq]
        power = power if type(power) is list else [power]
        param = param if type(param) is set() else list(param)
        param = param if type(param) is list else [param]

        # create figure
        fig = plt.figure()
        ax = fig.gca(projection='3d')

        surf = ax.plot([], [])
        fig.tight_layout()

        class Index(object):
            mesh = np.meshgrid(power, list(range(len(keys))))
            v = {'param': param, 'freq': freq, 'power': power}
            i = dict(zip(['param', 'freq'], [0, 0]))
            t = dict(zip(['param', 'freq'], [0, 0]))
            bbox_props = dict(boxstyle="round", fc="w", ec="0.5", alpha=0.9, )
            ref = {'fig': fig, 'plot': surf, 'axis': ax}
            ports = keys
            parent = self

            def _cb_rotate(self, event, direction, idx):
                self.i[idx] = np.mod(self.i[idx] + direction, len(self.v[idx]))
                self.redraw()

            def redraw(self):
                cparam, cfreq, cpower = [self.v['param'][self.i['param']], self.v['freq'][self.i['freq']], self.v['power']]

                (X, Y) = self.mesh
                Z = np.zeros(shape=(X.shape[0], X.shape[1]))
                for col, c in enumerate(Z):
                    for row, r in enumerate(c):
                        xt, val = self.parent.extractValues(cparam, freq=cfreq, power=cpower[row], keys=[self.ports[col]])
                        Z[col][row] = val.values[0]

                self.ref['axis'].cla()
                self.t['param'] = self.ref['axis'].text2D(0.05, 0.95, cparam, transform=ax.transAxes, bbox=self.bbox_props, picker=True)
                self.t['freq'] = self.ref['axis'].text2D(0.05, 0.90, cfreq, transform=ax.transAxes, bbox=self.bbox_props, picker=True)
                self.ref['axis'].plot_surface(X, Y, Z, cmap=cm.coolwarm,
                                              linewidth=1, antialiased=True, alpha=0.75, vmin=-174, vmax=-156, edgecolor='none')

                self.ref['axis'].set_title("surface plot : {}".format(cparam))
                self.ref['axis'].set_yticks(range(len(self.ports)))
                self.ref['axis'].set_yticklabels([str(x) for x in self.ports])
                self.ref['axis'].relim()
                self.ref['axis'].autoscale_view()
                self.ref['fig'].canvas.draw_idle()
                pass

        callback = Index()
        callback.redraw()

        def onpick1(event):
            if isinstance(event.artist, Text):
                text = event.artist
                direction = 1 if event.mouseevent.button == 1 or event.mouseevent.button == 'up' else -1

                if callback.t['param'] == text:
                    callback._cb_rotate(event=event, direction=direction, idx='param')
                elif callback.t['freq'] == text:
                    callback._cb_rotate(event=event, direction=direction, idx='freq')
                else:
                    pass

        cid = fig.canvas.mpl_connect('pick_event', onpick1)
        handler = (cid, callback)
        return handler

# ============================================================================


if __name__ == "__main__":
    pass



import unittest
import numpy as np
import rf_linkbudget as rf


class TestComponents(unittest.TestCase):

    def setUp(self):
        pass

    def test_VerifyParameterNumListOfTuples(self):
        cls = rf.VerifyParameterType().verify([(0, 0)])
        self.assertEqual(cls, rf.VerifyParameterNumListOfTuples)

    def test_VerifyParameterNumList(self):
        cls = rf.VerifyParameterType().verify([0, 0])
        self.assertEqual(cls, rf.VerifyParameterNumList)

    def test_VerifyParameterNumListSingleEntry(self):
        cls = rf.VerifyParameterType().verify([0])
        self.assertEqual(cls, rf.VerifyParameterNumListSingleEntry)

    def test_VerifyParameterNumSingleValue(self):
        cls = rf.VerifyParameterType().verify(0)
        self.assertEqual(cls, rf.VerifyParameterNumSingleValue)


if __name__ == '__main__':
    unittest.main()



import numpy as np


class VerifyParameterType():
    @staticmethod
    def verify(value):
        for cls in __class__.__subclasses__():
            if cls.verify(value):
                return cls

        raise ValueError('Parameter malformed : {}'.format(value))


class VerifyParameterNumListOfTuples(VerifyParameterType):
    @staticmethod
    def verify(value):
        if type(value) == list:
            if len(value) >= 1:
                if type(value[0]) == tuple:
                    return True
        return False


class VerifyParameterNumList(VerifyParameterType):
    @staticmethod
    def verify(value):
        if type(value) == list:
            if len(value) > 1:
                if type(value[0]) != tuple:
                    return True
        return False


class VerifyParameterNumListSingleEntry(VerifyParameterType):
    @staticmethod
    def verify(value):
        if type(value) == list:
            if len(value) == 1:
                if type(value[0]) != tuple:
                    return True
        return False


class VerifyParameterNumSingleValue(VerifyParameterType):
    @staticmethod
    def verify(value):
        if type(value) == int or type(value) == float:
            return True
        return False


if __name__ == '__main__':
    print( VerifyParameterType.verify([0]) )
    print( VerifyParameterType.verify([(0, 10)]) )
    print( VerifyParameterType.verify([0, 10, 20]) )
    print( VerifyParameterType.verify(1) )
    # print( VerifyParameterType.verify('a') )



# from scipy import constants
from abc import ABCMeta, abstractmethod
import numpy as np
from collections import OrderedDict
import networkx as nx
import itertools
import types
import copy

try:
    from .rfmath import RFMath
    from .result import simResult
    from . import verify
except ImportError:
    from rfmath import RFMath
    from result import simResult
    import verify

# ============================================================================
# Circuit Elements


class Port:
    """
    Port abstracts an input or output of a device. We register a callback at a Port

    Attributes
    ----------
    parent : AbstractDevice
        parent object
    num : int
        number of the port in the component
    """

    def __init__(self, parent, num):
        """
        Port abstracts an input or output of a device. We register a callback at a Port

        Parameters
        ----------
        parent : AbstractDevice
            parent object
        num : int
            number of the port in the component
        """
        self.num = num
        self.parent = parent
        self.connected = None

    @property
    def name(self):
        """
        Name of the Port.
        Composed of the parent name and the port number
        """
        return self.parent.name + ' Port {}'.format(self.num)

    def __rshift__(self, other):
        """
        connects two ports together. Checks for loops.

        Parameters
        ----------
        self : Port
            port to connect from
        other : Port
            port to connect to

        Raises
        ------
        ValueError
            if the second parameter is not also a Port
        ValueError
            if the second parameter is from the same device : a loop
        ValueError
            if Port is already connected

        Examples
        --------
         >>> amp1['out'] >> amp2['in']
        """
        if not isinstance(other, Port):
            raise ValueError('error, peer must also be port')

        if other.parent == self.parent:
            raise ValueError('error, loopback detected, peer port must be from another device')

        if self.connected is not None:
            raise ValueError('error, port is already connected')

        if other == self:
            raise ValueError('can not be the same port')

        self.parent.net.add_weighted_edges_from([(self, other, 1)], Gain=0)
        self.connected = True
        other.connected = True

    def __repr__(self):
        return "{}".format(self.name)

    def isOpen(self):
        """
        checks if the Port is connected to another port

        Returns
        --------
        isOpen : bool
            True if the Port is connected
        """
        return self.connected is not None

    def regCallback(self, func):
        """
        Register callback function. Is used to connect an "external" callback function to the port.
        This function gets called before each simulation with the current power and frequency.
        [dBm] / [Hz]

        Parameters
        ----------
        func : function
            function which gets added to the port

        Notes
        -----
        The callback function has the parameter freq, power
        """
        self.callback_preSimulation = types.MethodType(func, self)

    def regCallback_preIteration(self, func):  # bad name... I should change it...
        """
        Register callback function. Is used to connect an "external" callback function to the port.
        This function gets called before each simulation iteration of the specific port with the current input data, input frequency and power.
        [dict] / [Hz] / [dBm].
        In contrast to the regCallback function registered, this function gets called before iterating over the device, rather than at the beginning of the simulation.

        Parameters
        ----------
        func : function
            function which gets added to the port

        Notes
        -----
        The callback function has the parameter data, freq, power
        """
        self.callback_preIteration = types.MethodType(func, self)

    def callback_preIteration(self, data, f, p):
        """
        The actual callback function
        The intention of use is to set as example the output frequency of a mixer depending on the input frequency of the device.

        Parameters
        ----------
        data : dictionary
            the input data of the port.
        f : float
            the input frequency of the port.
        power : float
            the input power of the port.
        """
        return {}

    def callback_preSimulation(self, f, p):
        """
        The actual callback function.
        The intention of use is to set as example the input power / frequency of the simulation at the beginning of the simulation

        Parameters
        ----------
        f : float
            the input frequency of the port.
        power : float
            the input power of the port.
        """

        '''callback function for monkey patching to set initial values'''
        return {}

    def _callback_preSimulation(self, f, p):
        """
        hidden callback function to initialise the data dict with usefull defaults

        Parameters
        ----------
        f : float
            the input frequency of the port.
        power : float
            the input power of the port.

        Returns
        -------
        data : dictionary
            the output data
        """
        data = self.callback_preSimulation(f, p)
        if(len(data) > 0):
            for key, value in RFMath.defaults.items():
                if key not in data:
                    data[key] = value
            data.update(AbstractDevice.calcAdditionalParameters(data))

        return data

# ============================================================================


class Circuit:
    """
    Our base of the simulation is the Circuit class.
    We populate it with devices and make connections between them.
    Afterwards we simulate the circuit.

    Parameters
    ----------
    name : str
        circuit name
    """

    current = None
    """
    Circuit : self reference for easy circuit construction
    """

    currentNet = None
    """
    Networkx : temporary callback reference for the current used network in a simulation. Only used in the function Circuit._simulate(...)
    """

    currentSimParams = None
    """
    temporary callback reference for the current used parameters in a simulation. Only used in the function Circuit._simulate(...)
    """

    def register(child):
        """
        A function to register an AbstractDevice to the currently active Circuit

        Parameters
        ----------
        child : AbstractDevice
            a device to register to current active Circuit
        """
        Circuit.current.children.add(child)

    def __init__(self, name):
        self.name = name
        self.children = set()
        self.net = None
        Circuit.current = self

    def __getitem__(self, key):
        """
        convenience function to access named children (AbstractDevice)

        Parameters
        ----------
        key : str
            key to select a child (AbstractDevice)

        Returns
        -------
        val : AbstractDevice
            the item with key as name
        """
        return next(filter(lambda e: e.name == key, self.children))

    def __iter__(self):
        """
        convenience function to iterate over children names

        Returns
        -------
        iter : iterator
            creates iterator with all children names
        """
        return map(lambda e: e.name, self.children)

    def finalise(self):
        """
        do all checks before calculation, sweeps etc...

        Returns
        -------
        net : network
            returns generated and checked network
        """
        # Port.check()
        # Subcircuit.check()
        # Circuit.check()
        self.net = self.generateNetwork()
        return self.net

    def generateNetwork(self):
        """
        generate a network from all the AbstractDevice's which are registered at the current circuit

        Returns
        -------
        net : network
            returns generated network
        """
        net = nx.MultiDiGraph()
        [net.add_nodes_from(child.net.nodes(data=True)) for child in self.children]
        # ugly but copies attribute as well...
        [net.add_edge(edge[0], edge[1], **child.net[edge[0]][edge[1]][0]) for child in self.children for edge in child.net.edges]
        return net

    def simulate(self, network, start, end, freq, power, **kwargs):
        """
        simulate network

        Parameters
        ----------
        network : Network
            key to select a child (AbstractDevice)
        start : Port or AbstractDevice
            Starting Port of analysis
        end : Port or AbstractDevice
            Ending Port of analysis
        freq : list of floats
            System Sweep Frequency
        power : list of floats
            System Sweep Power

        Returns
        -------
        sim : rf_linkbudget.simResult
              returns simulation result
        """
        if type(start) != Port:
            start = start['out']  # assume the out port
        if type(end) != Port:
            end = end['in']  # assume the in port

        net = OrderedDict([(x, OrderedDict([(y, 0) for y in power])) for x in freq])  # initialise nested OrderedDict
        data = OrderedDict([(x, OrderedDict([(y, 0) for y in power])) for x in freq])

        for f, p in itertools.product(freq, power):
            n, d = self._simulate(network, start, end, f, p)
            net[f][p] = n
            data[f][p] = d

        return simResult(network=net, data=data)

    def _simulate(self, network, start, end, freq, power):
        """
        hidden simulate function. Which does the iterateive part of the simulation

        Parameters
        ----------
        network : Network
            key to select a child (AbstractDevice)
        start : Port or AbstractDevice
            Starting Port of analysis
        end : Port or AbstractDevice
            Ending Port of analysis
        freq : float
            System Sweep Frequency
        power : float
            System Sweep Power

        Returns
        -------
        (network, data) : (networkx, dictionary)
            returns simulation result
        """
        # 1) copy network and initialize working copy
        net = network.copy()
        Circuit.currentNet = net  # make a callback reference
        Circuit.currentSimParams = {'start': start, 'end': end, 'freq': freq, 'power': power}

        # 2) get all path related classes and call _callback_preSimulation before calc path
        [p._callback_preSimulation(freq, power) for p in net.nodes if hasattr(p.parent, 'updatePath')]

        # 3) get shortest path between start and end "port"
        ports = nx.shortest_path(net, start, end, weight='weight')

        # 4) create data structure
        data = OrderedDict([(p, {}) for p in ports])

        # 5) init data structure
        for key in data:
            data[key] = key._callback_preSimulation(freq, power)

        # 6) goto all ports and calculate the next ports value
        # decide if the edge is just a wire, then copy the parameters, otherwise compute
        for (s, e) in zip(ports[:-1], ports[1:]):
            if s.parent == e.parent:
                d = copy.deepcopy(data[s])
                d.update(s.callback_preIteration(d, freq, power))
                data[e] = s.parent.calcCurrentEdge(s, e, d)
                data[e].update(AbstractDevice.calcAdditionalParameters(data[e]))
            else:
                data[e] = data[s]  # copy data from one port to another, because its a wire...

        Circuit.currentNet = None  # clear callback reference
        Circuit.currentSimParams = None
        return (network, data)

# ============================================================================
# Abstracted Classes / Interfaces


class AbstractDevice:
    """
    AbstractDecices class is the base-class of all "Devices"


    Parameters
    ----------
    name : str
        circuit name
    n_ports : int
        Port count of the device
    **kwargs : undef
        additional Variables
    """
    __metaclass__ = ABCMeta

    def __init__(self, name, n_ports, **kwargs):
        self.name = name
        self.ports = [Port(self, num) for num in range(0, n_ports)]
        self.net = nx.MultiDiGraph()
        Circuit.register(self)

        for key, value in kwargs.items():
            setattr(self, key, value)

    def __getitem__(self, key):
        """
        convenience function to access the ports with numbers and strings

        Parameters
        ----------
        key : str or int
            key to select a Port

        Returns
        -------
        val : Port
            the item with key as name or as number
        """
        if type(key) != int:
            return self.ports[self._portkey[key]]
        else:
            return self.ports[key]

    @abstractmethod
    def calcCurrentEdge(self, start, end, data):
        """
        Abstract Function for calculation of the output data values

        Parameters
        ----------
        start : network edge (a Port)
            starting Port for the calculation
        end : network edge (a Port)
            ending Port for the calculation
        data : dictionary
            input data

        Returns
        -------
        data : dictionary
            calculated output data
        """
        raise NotImplementedError()

    def calcAdditionalParameters(data):
        """
        | Calc additional Parameters
        | Every AbstractDevice has to calculate its basic values like power, IP3, P1dB, etc.
        | In this function all other dependend values are calculated:

            - IM3D
            - IM3
            - n
            - NF
            - SNR
            - SFDR
            - DYN

        Parameters
        ----------
        data : dictionary
            input data

        Returns
        -------
        data : dictionary
            calculated output data
        """
        data['IMD3'] = RFMath.calc_IMD3(data['p'], data['IP3'])       # calc output IMD3 (simple, no cumulative content)
        data['IM3'] = RFMath.calc_IM3(data['p'], data['IP3'])         # calc output IM3 (simple, no cumulative content)
        data['n'] = RFMath.convert_T_n(data['Tn'])                    # Noise Power at 25°C degrees 1Hz Bandwidth
        data['NF'] = RFMath.convert_T_NF(data['Tn'], data['Gain'])    # NoiseFigure
        data['SNR'] = RFMath.calc_SNR(data['p'], data['n'])           # Signal to Noise Ration
        data['SFDR'] = RFMath.calc_SFDR(data['p'], data['IM3'])       # Spurious Free Dynamic Range
        data['DYN'] = RFMath.calc_Dynamic(data['SNR'], data['SFDR'])  # Dynamic
        return data

    def interpolateIfListOfFreqValTuple(self, freq, value):
        """
        if value is of the form: [(f,val)] then interpolate, otherwise ignore this function

        Parameters
        ----------
        freq : np.float
            frequency of interest
        value : float, or [(f,val)]
            value which might get interpolated

        Returns
        -------
        data : np.float
            value or interpolated value
        """

        if type(value) != list:
            return value
        if len(value) == 0:
            raise ValueError('value does not contain any values... thats shouldn\'t happen')
        if type(value[0]) == tuple:
            # Good thats what we expect and want to interpolate
            return np.interp(freq, *zip(*value))

# ============================================================================ #


class genericOnePort(AbstractDevice):
    """
    | A "generic" one port device.

    Parameters
    ----------
    name : string
        name of the device
    """

    def __init__(self, name):
        """
        Parameters
        ----------
        name : string
            name of the device
        """

        # Todo : implement checks and conversion to numpy arrays
        super().__init__(name, 1)
        self._portkey = {'in': 0, 'out': 0}
        self.net.add_nodes_from([port for port in self.ports])
        [self.net.add_edge(self.ports[0], port, weight=0) for port in self.ports[1:]]

# ============================================================================ #


class genericTwoPort(AbstractDevice):
    """
    | A "generic" two port device.

    Parameters
    ----------
    name : str
        device name
    Gain : list of [f, Gain]
        Gain representation in [dB], frequency in [Hz]
    Tn : list of [Tn @ Port 1, Tn @ Port 2] : Tn np.float or list of [f, Tn]
        noisetemperature of object, represented at the "target" port [°K]
    P1 : list of [P1 @ Port 1, P1 @ Port 2]
        Signal Compression Point of object, represented at the "target" port in [dB]
    IP3 : list of [P3 @ Port 1, P3 @ Port 2]
        Signal Intermodulation Point 3 of object, represented at the "target" port in [dB]
    """

    def __init__(self, name, Gain, Tn, P1, IP3):
        """
        Parameters
        ----------
        name : str
            device name
        Gain : list of [f, Gain]
            Gain representation in [dB], frequency in [Hz]
        Tn : list of [Tn @ Port 1, Tn @ Port 2]
            noisetemperature of object, represented at the "target" port [°K]
        P1 : list of [P1 @ Port 1, P1 @ Port 2]
            Signal Compression Point of object, represented at the "target" port in [dB]
        IP3 : list of [P3 @ Port 1, P3 @ Port 2]
            Signal Intermodulation Point 3 of object, represented at the "target" port in [dB]
        """
        # Todo : implement checks and conversion to numpy arrays
        super().__init__(name, 2, Gain=Gain, Tn=Tn, P1=P1, IP3=IP3)
        self._portkey = {'in': 0, 'out': 1}
        self.net.add_nodes_from([port for port in self.ports])
        [self.net.add_edge(self.ports[0], port, weight=0) for port in self.ports[1:]]

    @classmethod
    def fromSParamFile(cls, name, filename, Tn, P1, IP3, patchString='S12DB'):
        """
        | classmethod to create a two port device from a Touchstone S2P file
        | only S21 is regarded

        Parameters
        ----------
        name : str
            device name
        filename : str
            S2P filename
        Tn : list of [Tn @ Port 1, Tn @ Port 2]
            noisetemperature of object, represented at the "target" port [°K]
        P1 : list of [P1 @ Port 1, P1 @ Port 2]
            Signal Compression Point of object, represented at the "target" port in [dB]
        IP3 : list of [P3 @ Port 1, P3 @ Port 2]
            Signal Intermodulation Point 3 of object, represented at the "target" port in [dB]

        Other Parameters
        ----------------
        patchString : str
            default 'S12DB'

        Returns
        -------
        cls : genericTwoPort
            object
        """
        import skrf as rf

        sparam = rf.touchstone.Touchstone(filename)
        sparam = sparam.get_sparameter_data(format='db')

        Gain = [(f, s21) for f, s21 in zip(sparam['frequency'], sparam[patchString])]

        return cls(name, Gain=Gain, Tn=Tn, P1=P1, IP3=IP3)

    def calcCurrentEdge(self, start, end, data):
        """
        | calculates the output data values of a two port device
        | only the following values have to calculated inside this function:

            - f out
            - Gain
            - Tn out
            - p out
            - OP1dB
            - OIP3

        Parameters
        ----------
        start : network edge (a Port)
            starting Port for the calculation
        end : network edge (a Port)
            ending Port for the calculation
        data : dictionary
            input data

        Returns
        -------
        data : dictionary
            calculated output data
        """
        out = dict(data)                                                 # copy input values to output (as good defaults)

        if start == self.ports[0] and end == self.ports[1]:

            # Gain = np.interp(data['f'], *zip(*self.Gain))                 # interpolate Gain value
            Gain = self.interpolateIfListOfFreqValTuple(data['f'], self.Gain)
            Tn = self.interpolateIfListOfFreqValTuple(data['f'], self.Tn[1])
            P1 = self.interpolateIfListOfFreqValTuple(data['f'], self.P1[1])
            IP3 = self.interpolateIfListOfFreqValTuple(data['f'], self.IP3[1])

            out['Gain'] = data['Gain'] + Gain                             # calc cumulative Gain
            out['Tn'] = RFMath.calc_Tn(data['Tn'], Tn, Gain)              # calc output NoiseTemperature
            out['p'] = RFMath.calc_Pout(data['p'], Gain)                  # calc output signal power
            out['P1'] = RFMath.calc_P1(data['P1'], P1, Gain)              # calc output P1dB point
            out['IP3'] = RFMath.calc_IP3(data['IP3'], IP3, Gain)          # calc output IP3 point (simple)
            # calc other values...
            return out

        else:
            raise ValueError('this should not happen')

# ============================================================================ #


class Source(genericOnePort):
    """
    | A source device.

    Parameters
    ----------
    name : string
        name of the device

    Examples
    -------
    >>> src = Source("Source")
    >>> src['out'] >> amp1['in']
    """

    def __init__(self, name):
        super().__init__(name)

# ============================================================================ #


class Sink(genericOnePort):
    """
    | A sink device.

    Parameters
    ----------
    name : string
        name of the device

    Examples
    -------
    >>> snk = Sink("Sink")
    >>> amp2['out'] >> snk['in']
    """

    def __init__(self, name):
        super().__init__(name)

# ============================================================================ #


class Amplifier(genericTwoPort):
    """
    | An amplifier device.

    Parameters
    ----------
    name : str
        device name
    Gain : list of [f, Gain]
        Gain representation in [dB], frequency in [Hz]
    NF : numpy.float
        NoiseFigure in [dB]
    OP1dB : numpy.float
        Output Signal Compression Point in [dB]
    OIP3 : numpy.float
        Output Signal Intermodulation Point 3 in [dB]
    """

    def __init__(self, name, Gain, NF, OP1dB, OIP3):
        """
        Parameters
        ----------
        name : str
            device name
        Gain : list of [f, Gain]
            Gain representation in [dB], frequency in [Hz]
        NF : numpy.float
            NoiseFigure in [dB]
        OP1dB : numpy.float
            Output Signal Compression Point of object in [dB]
        OIP3 : numpy.float
            Output Signal Intermodulation Point 3 of object in [dB]
        """
        # Todo : implement checks and conversion to numpy arrays
        Tn = [0, 10**(NF/10) * RFMath.T0 - RFMath.T0]
        super().__init__(name, Gain=Gain, Tn=Tn, P1=[0, OP1dB], IP3=[0, OIP3])

    @classmethod
    def fromSParamFile(cls, name, filename, NF, OP1dB, OIP3, patchString='S12DB'):
        """
        | classmethod to create an amplifier device from a Touchstone S2P file
        | only S21 is regarded

        Parameters
        ----------
        name : str
            device name
        filename : str
            S2P filename
        NF : numpy.float
            NoiseFigure in [dB]
        OP1dB : numpy.float
            Output Signal Compression Point of object in [dB]
        OIP3 : numpy.float
            Output Signal Intermodulation Point 3 of object in [dB]

        Other Parameters
        ----------------
        patchString : str
            default 'S12DB'

        Returns
        -------
        cls : genericTwoPort
            object
        """
        import skrf as rf

        sparam = rf.touchstone.Touchstone(filename)
        sparam = sparam.get_sparameter_data(format='db')

        Gain = [(f, s21) for f, s21 in zip(sparam['frequency'], sparam[patchString])]

        return cls(name, Gain=Gain, NF=NF, OP1dB=OP1dB, OIP3=OIP3)

# ============================================================================ #


class Attenuator(genericTwoPort):
    """
    | An attenuator device.

    Parameters
    ----------
    name : str
        device name
    Att : numpy.float or [numpy.float]
        Attenuation representation in [dB]
    OP1dB : numpy.float
        Output Signal Compression Point in [dB]
    IIP3 : numpy.float
        Output Signal Intermodulation Point 3 in [dB]
    """

    def __init__(self, name, Att, OP1dB=None, IIP3=None):
        """
        Parameters
        ----------
        name : str
            device name
        Att : numpy.float or [numpy.float]
            Attenuation representation in [dB]
        OP1dB : numpy.float
            Output Signal Compression Point in [dB]
        IIP3 : numpy.float
            Input Signal Intermodulation Point 3 in [dB]
        """

        if type(Att) != int:
            _Att = Att
            Att = min(Att)

        # Todo : implement checks and conversion to numpy arrays
        OIP3 = IIP3 - Att if IIP3 is not None else None
        Tn = [0, 10**(Att/10) * RFMath.T0 - RFMath.T0]
        super().__init__(name, Gain=[(0, -Att)], Tn=Tn, P1=[0, OP1dB], IP3=[0, OIP3])
        self._Att = _Att if type(Att) != int else None
        self._IIP3 = IIP3

        for port in self.ports:
            setattr(port, 'setAttenuation', port.parent.setAttenuation)  # create port attribute and lin parent function to it

    def setAttenuation(self, Att):
        """
        | this setter is called in a callback function
        | the values gets rounded to the nearest value inside the Att parameter given in the construction of the object

        Parameters
        ----------
        Att : numpy.float
            Attenuation representation in [dB]
        """

        if len(self._Att) > 0:
            self._Att = np.array(self._Att)
            Att = self._Att[(np.abs(self._Att - Att)).argmin()]

        self.Tn = [0, 10**(Att/10) * RFMath.T0 - RFMath.T0]
        self.Gain = [(0, -Att)]

        if self._IIP3 is not None:
            self.IP3 = [0, self._IIP3 - Att]

# ============================================================================ #


class Filter(genericTwoPort):
    """
    | A filter device.

    Parameters
    ----------
    name : str
        device name
    Att : list of (f, Att)
        Attenuation representation in [dB]
    """

    def __init__(self, name, Att, OP1dB=None):
        """
        Parameters
        ----------
        name : str
            device name
        Att : list of (f, Att)
            Attenuation representation in [dB]
        OP1dB : numpy.float
            Output Signal Compression Point in [dB]
        """

        if not verify.VerifyParameterNumListOfTuples.verify(Att):
            raise ValueError('Parameter Att must be in the form [(freq, att), (freq, att)...]')
        if OP1dB:
            if verify.VerifyParameterNumListSingleEntry.verify(OP1dB):
                OP1dB = OP1dB[0]
            if not verify.VerifyParameterNumSingleValue.verify(OP1dB):
                raise ValueError('Parameter OP1dB must be in the form: val or [val]')

        Tn = [(f, 10**(att/10) * RFMath.T0 - RFMath.T0) for f, att in Att]
        Gain = [(f, -att) for f, att in Att]
        super().__init__(name, Gain=Gain, Tn=[0, Tn], P1=[0, OP1dB], IP3=[0, None])

    @classmethod
    def fromSParamFile(cls, name, filename, OP1dB=None, patchString='S12DB'):
        """
        | classmethod to create a filter device from a Touchstone S2P file
        | only S21 is regarded

        Parameters
        ----------
        name : str
            device name
        filename : str
            S2P filename
        OP1dB : numpy.float
            Output Signal Compression Point of object in [dB]

        Other Parameters
        ----------------
        patchString : str
            default 'S12DB'

        Returns
        -------
        cls : genericTwoPort
            object
        """
        import skrf as rf

        sparam = rf.touchstone.Touchstone(filename)
        sparam = sparam.get_sparameter_data(format='db')

        Att = [(f, -s21) for f, s21 in zip(sparam['frequency'], sparam[patchString])]

        return cls(name, Att=Att, OP1dB=OP1dB)


# ============================================================================ #


class SPDT(AbstractDevice):
    """
    | An spdt device.

    Parameters
    ----------
    name : str
        device name
    Att : numpy.float
        Attenuation representation in [dB]
    OP1dB : numpy.float
        Output Signal Compression Point in [dB]
    OIP3 : numpy.float
        Output Signal Intermodulation Point 3 in [dB]
    Iso : numpy.float
        Attenuation in case of "isolation" [dB]
    """

    def __init__(self, name, Att, OP1dB=None, OIP3=None, Iso=999):
        """
        Parameters
        ----------
        name : str
            device name
        Att : numpy.float
            Attenuation representation in [dB]
        OP1dB : numpy.float
            Output Signal Compression Point in [dB]
        OIP3 : numpy.float
            Output Signal Intermodulation Point 3 in [dB]
        Iso : numpy.float
            Attenuation in case of "isolation" [dB]
        """
        Tn = [0, 10**(Att/10) * RFMath.T0 - RFMath.T0]

        super().__init__(name, 3, Gain=[(0, -Att)], Tn=Tn, P1=[0, OP1dB, OP1dB], IP3=[0, OIP3, OIP3], Iso=Iso)
        self._portkey = {'S': 0, 'S-1': 1, 'S-2': 2}
        self.net.add_nodes_from([port for port in self.ports])
        [self.net.add_edge(self.ports[0], port, weight=0) for port in self.ports[1:]]
        [self.net.add_edge(port, self.ports[0], weight=0) for port in self.ports[1:]]
        self.setDirection(ref=self.net)

        for port in self.ports:
            setattr(port, 'setDirection', port.parent.setDirection)  # create port attribute and lin parent function to

        self.updatePath = True  # signals that this class has a function which is path dependend!

    def setDirection(self, dir=1, ref=None):
        """
        | gets called in pre-Simulation callback function.

        Parameters
        ----------
        name : str
            device name
        Att : numpy.float
            Attenuation representation in [dB]
        OP1dB : numpy.float
            Output Signal Compression Point in [dB]
        OIP3 : numpy.float
            Output Signal Intermodulation Point 3 in [dB]

        Notes
        -----
        | The implementation is a bit of a quirk.
        | we need a reference to the current net, which was implemented by **Circuit.currentNet**
        | but we need to delete the reference at the end of the simulation.
        | (end of **Circuit._simulate()** )

        | The switch has 3 Ports :

            - 'S'
            - 'S-1'
            - 'S-2'

        | The Ports can be used in any direction.

        Examples
        --------
        >>> sw1 = SPDT('switch 1')
        >>> src['out'] >> sw1['S']
        >>> sw1['S-1'] >> amp1['in']
        >>> sw1['S-2'] >> amp2['in']
        """
        dir = self._portkey[dir] if type(dir) != int else dir
        ref = Circuit.currentNet if ref is None else ref

        if dir == 1:
            ref.add_edge(self.ports[0], self.ports[1], key=0, weight=1)
            ref.add_edge(self.ports[1], self.ports[0], key=0, weight=1)
            ref.add_edge(self.ports[0], self.ports[2], key=0, weight=10)
            ref.add_edge(self.ports[2], self.ports[0], key=0, weight=10)
        elif dir == 2:
            ref.add_edge(self.ports[0], self.ports[1], key=0, weight=10)
            ref.add_edge(self.ports[1], self.ports[0], key=0, weight=10)
            ref.add_edge(self.ports[0], self.ports[2], key=0, weight=1)
            ref.add_edge(self.ports[2], self.ports[0], key=0, weight=1)
        else:
            raise ValueError()

    def calcCurrentEdge(self, start, end, data):
        """
        | calculates the output data values of an spdt device
        | only the following values have to calculated inside this function:

            - f out
            - Gain
            - Tn out
            - p out
            - OP1dB
            - OIP3

        Parameters
        ----------
        start : network edge (a Port)
            starting Port for the calculation
        end : network edge (a Port)
            ending Port for the calculation
        data : dictionary
            input data

        Returns
        -------
        data : dictionary
            calculated output data
        """
        out = dict(data)                                              # copy input values to output (as good defaults)

        if Circuit.currentNet.get_edge_data(start, end)[0]['weight'] > 1:
            # here we have the case with a switch in isolation state
            iso = self.interpolateIfListOfFreqValTuple(data['f'], self.Iso)
            Gain = -iso
        else:
            Gain = self.interpolateIfListOfFreqValTuple(data['f'], self.Gain)  # interpolate Gain value

        out['Gain'] = data['Gain'] + Gain                             # calc cumulative Gain
        out['Tn'] = RFMath.calc_Tn(data['Tn'], self.Tn[1], Gain)      # calc output NoiseTemperature
        out['p'] = RFMath.calc_Pout(data['p'], Gain)                  # calc output signal power
        out['P1'] = RFMath.calc_P1(data['P1'], self.P1[1], Gain)      # calc output P1dB point
        out['IP3'] = RFMath.calc_IP3(data['IP3'], self.IP3[1], Gain)  # calc output IP3 point (simple)
        # calc other values...
        return out

# ============================================================================ #


class Mixer(genericTwoPort):
    """
    | A Mixer device.

    Parameters
    ----------
    name : str
        device name
    Gain : list of [f, Gain]
            Gain representation in [dB], frequency in [Hz]
    OP1dB : numpy.float
        Output Signal Compression Point in [dB]
    OIP3 : numpy.float
        Output Signal Intermodulation Point 3 in [dB]

    Notes
    -----
    use callback_preIteration, to calc output frequency {'f': ...} the rest identical to a two port object

    Warnings
    --------
    not tested well at the moment
    """

    def __init__(self, name, Gain, OP1dB, OIP3, Tn=None):
        """
        Parameters
        ----------
        name : str
            device name
        Gain : list of [f, Gain]
                Gain representation in [dB], frequency in [Hz]
        OP1dB : numpy.float
            Output Signal Compression Point in [dB]
        OIP3 : numpy.float
            Output Signal Intermodulation Point 3 in [dB]

        Notes
        -----
        use callback_preIteration, to calc output frequency {'f': ...} the rest identical to a two port object
        """
        # Todo : implement checks and conversion to numpy arrays
        Tn = [(f, 10**(-g/10) * RFMath.T0 - RFMath.T0) for f, g in Gain] if Tn is None else Tn
        super().__init__(name, Gain=Gain, Tn=Tn, P1=[0, OP1dB], IP3=[0, OIP3])

    def calcCurrentEdge(self, start, end, data):
        """
        | calculates the output data values of an spdt device
        | only the following values have to calculated inside this function:

            - f out
            - Gain
            - Tn out
            - p out
            - OP1dB
            - OIP3

        Parameters
        ----------
        start : network edge (a Port)
            starting Port for the calculation
        end : network edge (a Port)
            ending Port for the calculation
        data : dictionary
            input data

        Returns
        -------
        data : dictionary
            calculated output data
        """
        out = dict(data)                                              # copy input values to output (as good defaults)

        freq = data['f'] if 'f_ref' not in data else data['f_ref']

        Gain = np.interp(freq, *zip(*self.Gain))                 # interpolate Gain value
        Tn = np.interp(freq, *zip(*self.Tn))                     # interpolate Gain value

        out['Gain'] = data['Gain'] + Gain                             # calc cumulative Gain
        out['Tn'] = RFMath.calc_Tn(data['Tn'], Tn, Gain)              # calc output NoiseTemperature
        out['p'] = RFMath.calc_Pout(data['p'], Gain)                  # calc output signal power
        out['P1'] = RFMath.calc_P1(data['P1'], self.P1[1], Gain)      # calc output P1dB point
        out['IP3'] = RFMath.calc_IP3(data['IP3'], self.IP3[1], Gain)  # calc output IP3 point (simple)
        # calc other values...
        return out

# ============================================================================ #

# TODO: ADC class

# ============================================================================ #


if __name__ == "__main__":

    pass



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(name='rf-linkbudget',
                 version='1.1.6',
                 description='A simple rf-linkbudget calculation tool',
                 url='https://github.com/hwengineer/rf_linkbudget',
                 author='Alexander Ott',
                 author_email='alexander.ott.ottale@gmail.com',
                 long_description=long_description,
                 long_description_content_type="text/markdown",
                 license='MIT',
                 packages=setuptools.find_packages(),
                 classifiers=["Programming Language :: Python :: 3",
                              "License :: OSI Approved :: MIT License",
                              "Operating System :: OS Independent", ],
                 python_requires='>=3.5',
                 zip_safe=False)



from scipy import constants
import numpy as np
import unittest

# ============================================================================ #


class RFMath:
    """
    Calculation Class for all the different Parameters

    Note
    ----
    search in this class for calculation errors
    """

    T0 = np.float(290)
    """
    numpy.float : Noisetemperature at N0
    """
    N0 = np.float(10*np.log10(constants.k * 290 * 1 * 1000))
    """
    numpy.float : Noisefloor at 25°C and 50R
    """

    defaults = {'P1': None, 'IP3': None, 'Tn': 0, 'Gain': 0}
    """
    dict : Sane default values for result dictionary
    """

    def convert_T_n(T, B=1):
        """
        Convert Noise Temperature to Noise Power

        Parameters
        ----------
        T : float
            Noise Temperature in [°K]

        Returns
        -------
        n : numpy.float
            noise power in [dBm/Hz] / [dBm/(B*Hz)]
        """
        return 10*np.log10(constants.k * T * B * 1000) if T != 0 else -np.inf

    def convert_T_NF(T, Gain):
        """
        Convert Noise Temperature to NoiseFigure

        Parameters
        ----------
        T : float
            Noise Temperature in [°K]

        Gain : float
            Accumulated Gain in [dB]

        Returns
        -------
        NF : numpy.float
            Noisefigure in [dB]

        Notes
        -----
        .. _Wiki: https://en.wikipedia.org/wiki/Noise_temperature
        """
        g = 10**(Gain/10)
        Teq = T/g
        F = Teq/RFMath.T0

        return 10*np.log10(F)

    def calc_IM3(P, IP3):
        """
        Calculate Intermodulation Value IM3 (absolute)
        from P and IP3

        Parameters
        ----------
        P : float
            Signal ower in [dBm]
        IP3 : float
            IP3 in [dBm]

        Returns
        -------
        IM3 : float or None
            intermodulation 3 power in [dBm]
        """
        return P - (2*(IP3 - P) + 6) if(IP3 is not None) else None

    def calc_IMD3(P, IP3):
        """
        Calculate Intermodulation 3 distance between P and IM3
        from P and IP3

        Parameters
        ----------
        P : float
            Signal ower in [dBm]
        IP3 : float
            IP3 in [dBm]

        Returns
        -------
        IMD3 : float or None
            distance between signal power and intermodulation power in [dBm]
        """
        return 2*(IP3 - P) + 6 if(IP3 is not None) else None

    def calc_SNR(P, n):
        """
        Calculate SNR value
        from P and n

        Parameters
        ----------
        P : float
            Signal power in [dBm]
        n : float
            Noise power in [dBm/Hz]

        Returns
        -------
        SNR : float
            distance between Signal and Noise power in [dB]
        """
        return P - n

    def calc_SFDR(P, IM3):
        """
        Calculate SFDR value
        from P and IM3

        Parameters
        ----------
        P : float
            Signal power in [dBm]
        IM3 : float
            Intermodulation 3 power in [dBm]

        Returns
        -------
        SFDR : float or None
            Spurious free dynamic range in [dB]
        """
        return P - IM3 if(IM3 is not None) else None

    def calc_Dynamic(SNR, SFDR):
        """
        Calculate Dynamic value
        from SNR and SFDR

        Parameters
        ----------
        SNR : float
            Signal to Noise ratio in [dB]
        SFDR : float
            Spurious free dynamic range in [dB]

        Returns
        -------
        Dynamic : int
            Dynamic, min(SNR, SFDR) in [dB]
        """
        return min(SNR, SFDR) if(SFDR is not None) else None

    def calc_P1(P1_in, P1, Gain):
        """
        Calculate P1 value

        Parameters
        ----------
        P1_in : float
            Compression Point 1 from predecessor stage in [dBm]
        P1: float
            Compression Point 1 in [dBm]
        Gain: float
            Gain in [dB]

        Returns
        -------
        P1 : float or None
            Output Compression Point 1 in [dBm]
        """

        v1 = P1_in + Gain if P1_in is not None else None
        v2 = P1
        return None if not any([v1, v2]) else min(filter(None, [v1, v2]))  # return None or minimum of both variables

    def calc_IP3(IP3_in, IP3, Gain):
        """
        Calculate IP3 value

        Parameters
        ----------
        IP3_in : float
            Compression Point 3 from predecessor stage in [dBm]
        IP3: float
            Compression Point 3 in [dBm]
        Gain: float
            Gain in [dB]

        Returns
        -------
        IP3 : float or None
            Output Compression Point 3 in [dBm]
        """

        v1 = IP3_in + Gain if IP3_in is not None else None
        v2 = IP3
        return None if not any([v1, v2]) else min(filter(None, [v1, v2]))  # return None or minimum of both variables

    def calc_Pout(P, Gain):
        """
        Calculate Signal output power

        Parameters
        ----------
        P : float
            Input signal power in [dBm]
        Gain: float
            Gain in [dB]

        Returns
        -------
        P : float
            Output signal power in [dBm]
        """
        return P + Gain

    def calc_Tn(Tn_in, Tn, Gain):
        """
        Calculate output noise temperature

        Parameters
        ----------
        Tn_in : float
            Input noise temperature in [°K]
        Tn : float
            Noise temperature of this stage in [°K]
        Gain: float
            Gain in [dB]

        Returns
        -------
        Tn : float
            Output noise temperature in [°K]
        """
        return 10**(Gain/10) * (Tn_in + Tn)

# ============================================================================ #


class RFMathTest(unittest.TestCase):
    def test_calc_SNR(self):
        value = RFMath.calc_SNR(10, 5)
        self.assertEqual(5, value)

        # assertEquals
        # assertTrue
        # assertFalse
        # assertRaises
    def test_calc_SFDR(self):
        value = RFMath.calc_SFDR(10, 5)
        self.assertEqual(5, value)

    def test_calc_Dynamic(self):
        value = RFMath.calc_Dynamic(10, 6)
        self.assertEqual(6, value)

    def test_calc_Dynamic2(self):
        value = RFMath.calc_Dynamic(3, 6)
        self.assertEqual(3, value)

# ============================================================================ #


if __name__ == "__main__":
    unittest.main()





from .circuit import Port
from .circuit import Circuit
from .circuit import AbstractDevice
from .circuit import genericOnePort
from .circuit import genericTwoPort

from .circuit import Source
from .circuit import Sink

from .circuit import Amplifier
from .circuit import Attenuator
from .circuit import Filter
from .circuit import SPDT
from .circuit import Mixer
#from .circuit import ADC

from .rfmath import RFMath

from .result import simResult


from .verify import * 
